/*-------------------------------------------------------------------------+
| Copyright 2018 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.editingsupport;

import static java.lang.Integer.compare;
import static java.lang.Math.min;
import static java.util.Arrays.asList;
import static java.util.Arrays.stream;

import java.util.Collection;
import java.util.OptionalInt;
import java.util.stream.IntStream;

import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SegmentEvent;
import org.eclipse.swt.events.SegmentListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Text;

/**
 * Dialog that allows to edit multiple values
 * 
 * @author barner
 */
public class MultiValueAnnotationTextEditingDialog<T> extends MultiValueAnnotationDialogBase<T> {

	/** The {@link Text} into which input is entered. */
	private Text textInput;

	/** {@link List} containing the current input. */
	private List list;

	/** {@link Button} to add the current value of {@link #textInput} to {@link #list}. */
	private Button addButton;

	/** {@link Button} to remove the currently selected item from {@link #list}. */
	private Button removeButton;

	/** {@link Button} to move up the currently selected item of {@link #list}. */
	private Button upButton;

	/** {@link Button} to move down the currently selected item of {@link #list}. */
	private Button downButton;

	/** Flag if dialog is currently set up in {@link #createDialogArea(Composite)}. */
	private boolean isCreateDialogArea;

	/** Underlying {@link EAttribute}. */
	private EAttribute eAttribute;

	/** The {@link Collection} of initial values. */
	private Collection<T> initialValues;

	/** Constructs a new {@link MultiValueAnnotationTextEditingDialog}. */
	public MultiValueAnnotationTextEditingDialog(EAttribute eAttribute,
			Collection<T> initialValues) {
		super("Add", eAttribute.getLowerBound(), eAttribute.getUpperBound());
		this.eAttribute = eAttribute;
		this.initialValues = initialValues;
	}

	/** Returns the {@link EFactory} for the underlying {@link #eAttribute}. */
	protected EFactory getEFactory() {
		return eAttribute.getEType().getEPackage().getEFactoryInstance();
	}

	/** Refreshes the status of the dialog's UI elements. */
	protected void refresh() {
		if(isCreateDialogArea) {
			return;
		}

		boolean isValidText = true;
		try {
			EFactory eFactory = getEFactory();
			eFactory.createFromString(eAttribute.getEAttributeType(), textInput.getText());
		} catch(Exception e) {
			isValidText = false;
		}

		addButton.setEnabled(isValidText);

		OptionalInt min = stream(list.getSelectionIndices()).reduce(Integer::min);
		OptionalInt max = stream(list.getSelectionIndices()).reduce(Integer::max);
		removeButton.setEnabled(min.isPresent());

		upButton.setEnabled(min.isPresent() && min.getAsInt() > 0);
		downButton.setEnabled(max.isPresent() && max.getAsInt() < list.getItemCount() - 1);
	}

	/** {@inheritDoc} */
	@Override
	protected Composite createDialogArea(Composite parent) {
		isCreateDialogArea = true;
		Composite area = (Composite)super.createDialogArea(parent);
		area.setLayout(new GridLayout(2, false));

		textInput = new Text(area, SWT.BORDER | SWT.LEFT);
		GridData gd_Text = new GridData(SWT.LEFT, SWT.CENTER, true, false, 1, 1);
		gd_Text.widthHint = 300;
		gd_Text.minimumWidth = 300;
		textInput.setLayoutData(gd_Text);
		textInput.addSegmentListener(new SegmentListener() {
			/** {@inheritDoc} */
			@Override
			public void getSegments(SegmentEvent event) {
				refresh();
			}
		});

		addButton = new Button(area, SWT.PUSH);
		GridData gd_Buttons = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		gd_Buttons.widthHint = 100;
		gd_Buttons.minimumWidth = 100;
		addButton.setLayoutData(gd_Buttons);
		addButton.setText("Add");

		addButton.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				list.add(textInput.getText());
				textInput.setText("");
				refresh();
			}
		});

		list = new List(area, SWT.BORDER | SWT.LEFT | SWT.MULTI);
		GridData gd_List = new GridData(SWT.LEFT, SWT.CENTER, true, false, 1, 4);
		gd_List.minimumWidth = 310;
		gd_List.heightHint = 310;
		list.setLayoutData(gd_List);
		list.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				refresh();
			}
		});
		for(Object value : initialValues) {
			EFactory eFactory = getEFactory();
			String valueLabel = eFactory.convertToString(eAttribute.getEAttributeType(), value);
			list.add(valueLabel);
		}

		removeButton = new Button(area, SWT.PUSH);
		removeButton.setLayoutData(gd_Buttons);
		removeButton.setText("Remove");
		removeButton.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				int[] selectionIndices = list.getSelectionIndices();
				int index = -1;
				if(selectionIndices.length == 1) {
					index = selectionIndices[0];
				}
				for(int i : selectionIndices) {
					list.remove(i);
				}
				list.setSelection(min(index, list.getItemCount()));
				refresh();
			}
		});

		upButton = new Button(area, SWT.PUSH);
		upButton.setLayoutData(gd_Buttons);
		upButton.setText("Up");
		upButton.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				stream(list.getSelectionIndices()).sorted().forEach(i -> {
					String swap = list.getItem(i - 1);
					list.setItem(i - 1, list.getItem(i));
					list.setItem(i, swap);
				});
				IntStream oldIndices = stream(list.getSelectionIndices());
				list.deselectAll();
				list.select(oldIndices.map(i -> i - 1).toArray());
				refresh();
			}
		});

		downButton = new Button(area, SWT.PUSH);
		downButton.setLayoutData(gd_Buttons);
		downButton.setText("Down");
		downButton.addSelectionListener(new SelectionAdapter() {
			/** {@inheritDoc} */
			@Override
			public void widgetSelected(SelectionEvent e) {
				stream(list.getSelectionIndices()).boxed().sorted((i1, i2) -> compare(i2, i1))
						.forEach(i -> {
							String swap = list.getItem(i + 1);
							list.setItem(i + 1, list.getItem(i));
							list.setItem(i, swap);
						});
				IntStream oldIndices = stream(list.getSelectionIndices());
				list.deselectAll();
				list.select(oldIndices.map(i -> i + 1).toArray());
				refresh();
			}
		});

		isCreateDialogArea = false;
		refresh();
		return area;
	}

	/** {@inheritDoc} */
	@Override
	public void create() {
		super.create();
		getShell().setDefaultButton(addButton);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	protected Collection<T> getElementsFromDialog() {
		return (Collection<T>)asList(list.getItems());
	}
}
