/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.addConnectionPoint;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.base.HierarchicElementBase;
import org.fortiss.tooling.base.model.element.IConnection;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.DelegatingContentAnchorageController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

/**
 * {@link DelegatingContentAnchorageController} for diagrams with curved links.
 * 
 * @author munaro
 */
public class CurvedLinkLayoutedContentAnchorangeController<T extends IConnector & INamedCommentedElement & ILayoutedModelElement>
		extends LayoutedContentAnchorageController<T> {

	/** Constructor. */
	public CurvedLinkLayoutedContentAnchorangeController(IContentAnchorageMVCBundle mvcb,
			Class<T> modelType) {
		super(mvcb, modelType);
	}

	/** {@inheritDoc} */
	// TODO(3978): Resolve code duplication in the link-method of link visuals
	@Override
	protected void link(IMVCBundle startBundle, DiagramCoordinate startLocation,
			IMVCBundle endBundle, DiagramCoordinate endLocation) {
		// Disable direct links between hierarchic element bases
		if(startBundle.getModel() instanceof HierarchicElementBase ||
				endBundle.getModel() instanceof HierarchicElementBase) {
			return;
		}

		// The call to `super.link` (see next lines) does not return a reference to the
		// newly created connection. However, as it is needed in order to add the handles
		// we get it by comparing the root element's connections before and after the
		// creation of the new link:
		// 1) Here we get the list of existing connections
		EList<IConnection> connectionsBefore = new BasicEList<IConnection>(
				((HierarchicElementBase)(EObject)getViewer().getRootElement()).getConnections());
		// 2) Here we create the new connection
		super.link(startBundle, startLocation, endBundle, endLocation);
		// 3) Here we get the updated list of connections
		EList<IConnection> connectionsAfter = new BasicEList<IConnection>(
				((HierarchicElementBase)(EObject)getViewer().getRootElement()).getConnections());
		// 4) We remove all existing ones
		connectionsAfter.removeAll(connectionsBefore);
		// 5) And we are left with the newly added one.
		IConnection connection = connectionsAfter.stream().findFirst().get();

		// Create handles
		DiagramCoordinate startPosition =
				new DiagramCoordinate(startBundle.getVisual().getCurrentBounds().getMaxX(),
						startBundle.getVisual().getCurrentBounds().getMaxY());
		DiagramCoordinate endPosition =
				new DiagramCoordinate(endBundle.getVisual().getCurrentBounds().getMaxX(),
						endBundle.getVisual().getCurrentBounds().getMaxY());
		DiagramCoordinate middle =
				new DiagramCoordinate((startPosition.getX() + endPosition.getX()) / 2,
						(startPosition.getY() + endPosition.getY()) / 2);
		addConnectionPoint((ILayoutedModelElement)connection, 0, (int)middle.getX(),
				(int)middle.getY());
		addConnectionPoint((ILayoutedModelElement)connection, 1, (int)middle.getX(),
				(int)middle.getY());
	}
}
