/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import java.util.ArrayList;
import java.util.EventObject;
import java.util.List;

import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.IModelChangeProvider;
import org.fortiss.tooling.common.ui.javafx.lwfxef.model.ModelChangeProviderBase;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Base class for {@link IModelChangeProvider}s using kernel services.
 * 
 * @author hoelzl
 */
abstract class KernelServiceBasedModelChangeProviderBase<T extends EObject>
		extends ModelChangeProviderBase implements CommandStackListener {
	/** The persistency service reference. */
	private static final IPersistencyService persistencyService = IPersistencyService.getInstance();
	/** The command stack service. */
	private static final ICommandStackService commandStackService =
			ICommandStackService.getInstance();
	/** List of relevant {@link Notification}s recorded during a change transaction. */
	private final List<Notification> notifications = new ArrayList<>();

	/** Returns the model element. */
	protected abstract T getModelElement();

	/** Returns the EMF notification listener. */
	protected abstract Adapter getNotificationListener();

	/** Adds the given {@link Notification} to list of notifications. */
	protected final void addNotification(Notification event) {
		notifications.add(event);
	}

	/** {@inheritDoc} */
	@Override
	protected final void connectToModel() {
		notifications.clear();
		T modelElement = getModelElement();
		Adapter notificationListener = getNotificationListener();
		ITopLevelElement root = persistencyService.getTopLevelElementFor(modelElement);
		commandStackService.addCommandStackListener(root, this);
		modelElement.eAdapters().add(notificationListener);
	}

	/** {@inheritDoc} */
	@Override
	protected final void disconnectFromModel() {
		T modelElement = getModelElement();
		Adapter notificationListener = getNotificationListener();
		modelElement.eAdapters().remove(notificationListener);
		ITopLevelElement root = persistencyService.getTopLevelElementFor(modelElement);
		commandStackService.removeCommandStackListener(root, this);
		notifications.clear();
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		if(notifications.isEmpty()) {
			return;
		}
		fireUpdateComplete();
		notifications.clear();
	}
}
