/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor.fx.controller;

import static java.util.Arrays.asList;
import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;

import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

import javafx.geometry.Point2D;

/**
 * {@link IController} for {@link ILayoutedModelElement}s representing curved links and connections
 * in a {@link DiagramViewer}.
 * 
 * @author munaro
 */
public class LayoutedCurveLinkBendPointController<T extends INamedCommentedElement & ILayoutedModelElement>
		extends LayoutedLinkBendPointController<T> {

	/** Constructor. */
	public LayoutedCurveLinkBendPointController(ILinkMVCBundle mvcb, Class<T> modelType) {
		super(mvcb, modelType);
	}

	/** {@inheritDoc} */
	@Override
	protected final void createBendPointAt(int bpIndex, DiagramCoordinate location) {
		DiagramViewerFeatures features = getViewer().getFeatures();
		Point c2 = createPoint((int)(location.getX() - 2 * features.getHorizontalSpacing()),
				(int)location.getY(), null);
		Point c1 = createPoint((int)(location.getX() + 2 * features.getHorizontalSpacing()),
				(int)location.getY(), null);
		Point bp = createPoint((int)location.getX(), (int)location.getY(), null);
		getBendPointList(getModelElement()).addAll(bpIndex - 1, asList(c2, bp, c1));
	}

	/** {@inheritDoc} */
	@Override
	protected final void moveBendPointVisually(int bpIndex, Point2D delta) {
		super.moveBendPointVisually(bpIndex, delta);
		if(bpIndex <= 0 || bpIndex >= getNumerOfBendPoints() - 1) {
			// first and last bend-point move unconnected
			return;
		}
		ILinkVisual linkVisual = getLinkVisual();
		FeedbackChange centerPointChange = getMoveDeltaFeedback(bpIndex, delta);
		int indexMod3 = bpIndex % 3;
		if(indexMod3 == 2) {
			// bend-points between curve segments also drag the previous and next point
			linkVisual.setFeedbackChangeForBendPoint(bpIndex - 1, centerPointChange);
			linkVisual.setFeedbackChangeForBendPoint(bpIndex + 1, centerPointChange);
		} else {
			// otherwise the connected bend-point is moved in the opposite direction
			FeedbackChange inv = locationFeedbackChange(-centerPointChange.getDeltaX(),
					-centerPointChange.getDeltaY());
			int otherIndex = (indexMod3 == 1) ? bpIndex + 2 : bpIndex - 2;
			linkVisual.setFeedbackChangeForBendPoint(otherIndex, inv);
		}
		// linkVisual.update();
		return;
	}

	/** {@inheritDoc} */
	@Override
	protected final void moveBendPointInModel(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta);
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
		updateModelAfterBendPointMove(bpIndex, chg.getDeltaX(), chg.getDeltaY());
		if(bpIndex <= 0 || bpIndex >= getNumerOfBendPoints() - 1) {
			// first and last bend-point move unconnected
			return;
		}
		ILinkVisual linkVisual = getLinkVisual();
		int indexMod3 = bpIndex % 3;
		if(indexMod3 == 2) {
			// bend-points between curve segments also drag the previous and next point
			linkVisual.setFeedbackChangeForBendPoint(bpIndex - 1, null);
			updateModelAfterBendPointMove(bpIndex - 1, chg.getDeltaX(), chg.getDeltaY());
			linkVisual.setFeedbackChangeForBendPoint(bpIndex + 1, null);
			updateModelAfterBendPointMove(bpIndex + 1, chg.getDeltaX(), chg.getDeltaY());
		} else {
			// otherwise the connected bend-point is moved in the opposite direction
			int otherIndex = (indexMod3 == 1) ? bpIndex + 2 : bpIndex - 2;
			linkVisual.setFeedbackChangeForBendPoint(otherIndex, null);
			updateModelAfterBendPointMove(otherIndex, -chg.getDeltaX(), -chg.getDeltaY());
		}
	}
}
