/*******************************************************************************
 * Copyright (c) 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.fortiss.tooling.base.ui.editor.fx.visual;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.base.ui.utils.LWFXEditorUtils.computeLinkToCircleLocation;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUIUtils.getConnectionPoints;

import java.util.List;

import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.base.model.layout.Points;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic.CurveLinkVisualBase;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;

/**
 * {@link CurveLinkVisualBase} providing a default visual for wires in ellipses-and-wires diagrams
 * based on {@link ILayoutedModelElement}s.
 * 
 * @author munaro
 */
public class LayoutedCurveLinkVisual<T extends ILayoutedModelElement> extends CurveLinkVisualBase {
	/** Constructor. */
	public LayoutedCurveLinkVisual(ILinkMVCBundle mvcb, Class<T> modelType) {
		super(mvcb);

		// TODO(#3877): Move type checks to a common base class.
		Object model = mvcb.getModel();
		if(model == null) {
			throw new IllegalArgumentException("The given model is null!");
		}
		if(!modelType.isAssignableFrom(model.getClass())) {
			throw new IllegalArgumentException("Expected model of type " +
					modelType.getSimpleName() + ", but was " + model.getClass().getSimpleName());
		}
	}

	/** Returns the model element with the expected type. */
	@SuppressWarnings("unchecked")
	protected T getModelElement() {
		return (T)getModel();
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getBendPointLocation(int bpIndex) {
		Point point = getBendPointList().get(bpIndex);
		return new DiagramCoordinate(point.getX(), point.getY());
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getStartAnchorLocation(Rectangle2D anchorBounds, Point2D target) {
		return computeLinkToCircleLocation(anchorBounds, target);
	}

	/** {@inheritDoc} */
	@Override
	protected DiagramCoordinate getEndAnchorLocation(Rectangle2D anchorBounds, Point2D target) {
		return computeLinkToCircleLocation(anchorBounds, target);
	}

	/** {@inheritDoc} */
	@Override
	protected Object getBendPointModel(int i) {
		return getBendPointList().get(i);
	}

	/** {@inheritDoc} */
	@Override
	protected int getNumberOfBendPoints() {
		return getBendPointList().size();
	}

	/** Returns the list of bend-points. */
	private List<Point> getBendPointList() {
		Points connectionPoints = getConnectionPoints(getModelElement());
		if(connectionPoints == null) {
			return emptyList();
		}
		List<Point> pointList = connectionPoints.getPoints();
		if(pointList == null) {
			return emptyList();
		}
		return pointList;
	}

	/** {@inheritDoc} */
	@Override
	protected double getFeedbackMarkerSize() {
		return 10;
	}

	/** {@inheritDoc} */
	@Override
	protected boolean showArrowOnLastSegment() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected double getInvisibleSelectionLineWidth() {
		return 7;
	}

	/** {@inheritDoc} */
	@Override
	protected double getArrowLength() {
		return 7;
	}
}
