/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.utils;

import static org.fortiss.tooling.base.utils.LayoutModelElementFactory.createPoint;
import static org.fortiss.tooling.kernel.service.IPrototypeService.PROTOTYPE_DATA_FORMAT;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.dnd.ElementDropContext;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IPrototypeService;

import javafx.scene.input.Dragboard;

/**
 * Utility methods for JavaFX drag-and-drop.
 * 
 * @author hoelzl
 */
public final class FXDNDUtils {
	/** Returns whether element composition can be done via {@link IElementCompositorService}. */
	public static boolean canCompose(Dragboard db, EObject container) {
		Prototype proto = extractPrototype(db);
		if(proto != null) {
			return IElementCompositorService.getInstance().canCompose(container,
					proto.getPrototypeCopy(), null);
		}
		return false;
	}

	/** Returns whether element composition can be done via {@link IElementCompositorService}. */
	public static boolean canCompose(Dragboard db, DiagramCoordinate location, EObject container,
			boolean isRootContainer, double zoom) {
		Prototype proto = extractPrototype(db);
		if(proto != null) {
			ElementDropContext ctx =
					createElementDropContext(location, container, isRootContainer, zoom);
			return IElementCompositorService.getInstance().canCompose(container,
					proto.getPrototypeCopy(), ctx);
		}
		return false;
	}

	/** Composes the elements using {@link IElementCompositorService}. */
	public static boolean compose(Dragboard db, EObject container) {
		Prototype proto = extractPrototype(db);
		if(proto != null) {
			EObject copy = proto.getPrototypeCopy();
			return doElementComposition(container, copy, null);
		}
		return false;
	}

	/**
	 * Composes the elements using {@link IElementCompositorService} creating an
	 * {@link ElementDropContext} for the given location, root element, and zoom factor.
	 */
	public static boolean compose(Dragboard db, DiagramCoordinate location, EObject container,
			boolean isRootContainer, double zoom) {
		Prototype proto = extractPrototype(db);
		if(proto != null) {
			EObject copy = proto.getPrototypeCopy();
			ElementDropContext ctx =
					createElementDropContext(location, container, isRootContainer, zoom);
			return doElementComposition(container, copy, ctx);
		}
		return false;
	}

	/** Delegates to {@link IElementCompositorService} in order to compose the elements. */
	private static boolean doElementComposition(EObject container, EObject contained,
			ElementDropContext context) {
		IElementCompositorService ecs = IElementCompositorService.getInstance();
		if(ecs.canCompose(container, contained, context)) {
			ICommandStackService.getInstance().runAsCommand(container, () -> {
				ecs.compose(container, contained, context);
			});
			return true;
		}
		return false;
	}

	/** Creates the {@link ElementDropContext} containing the drop location. */
	private static ElementDropContext createElementDropContext(DiagramCoordinate location,
			EObject container, boolean isRootContainer, double zoom) {
		Point dcLocation = createPoint((int)location.getX(), (int)location.getY(), "");
		ElementDropContext ctx =
				new ElementDropContext(container, dcLocation, isRootContainer, zoom);
		return ctx;
	}

	/** Extracts the prototype from the dragboard. */
	private static Prototype extractPrototype(Dragboard db) {
		Object data = db.getContent(PROTOTYPE_DATA_FORMAT);
		if(data instanceof String) {
			return IPrototypeService.getInstance().getPrototypeByUniqueID((String)data);
		}
		return null;
	}

	/** Constructor. */
	private FXDNDUtils() {
		// prevent instantiation
	}
}
