/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.viewers;

import static org.eclipse.core.runtime.Platform.getBundle;
import static org.eclipse.jface.resource.JFaceResources.getImageRegistry;
import static org.fortiss.tooling.base.utils.SystemUtils.isLinuxPlatform;
import static org.fortiss.tooling.base.utils.SystemUtils.isMacOSXPlatform;

import org.eclipse.jface.resource.ImageRegistry;
import org.eclipse.jface.resource.JFaceResources;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;

/**
 * Checkbox label provider using screenshots of native checkbox buttons as representation. This
 * class supports checked and enabled state.
 * 
 * @author hoelzl
 */
// Using drawn images may seem strange, but internet sources could not provide
// any other solution to having checkbox in cell editors
public abstract class CheckBoxLabelProvider extends ColumnLabelProvider {
	/** {@link JFaceResources} id for checked image. */
	private static final String CHECKED_KEY = "CHECKED";

	/** {@link JFaceResources} id for unchecked image. */
	private static final String UNCHECKED_KEY = "UNCHECKED";

	/** {@link JFaceResources} id for disabled, unchecked image. */
	private static final String DISABLED_UNCHECKED_KEY = "DISABLED_UNCHECKED";

	/** {@link JFaceResources} id for disabled, checked image. */
	private static final String DISABLED_CHECKED_KEY = "DISABLED_CHECKED";

	/** Predicate whether to use a text-based icon (see #4032 and #2443). */
	public static boolean useTextIcon() {
		return isMacOSXPlatform() || isLinuxPlatform();
	}

	/** Text label for unchecked checkboxes when {@link #useTextIcon()} returns {@code true}. */
	public static final String UNCHECKED_LABEL = "\u25A1";

	/** Text label for checked checkboxes when {@link #useTextIcon()} returns {@code true}. */
	public static final String CHECKED_LABEL = "\u2612";

	/** Creates an image of checkbox SWT control. */
	private static Image makeShot(boolean selected, boolean enabled) {
		Shell shell = new Shell(SWT.NO_TRIM);
		Display display = shell.getDisplay();

		// Hopefully no platform uses exactly this color because we'll make
		// it transparent in the image.
		Color greenScreen = new Color(display, 222, 223, 224);

		// Otherwise we have a default gray color
		shell.setBackground(greenScreen);

		Button button = new Button(shell, SWT.CHECK);
		button.setBackground(greenScreen);
		button.setSelection(selected);
		button.setEnabled(enabled);

		// Otherwise an image is located in a corner
		button.setLocation(1, 1);
		Point bsize = button.computeSize(SWT.DEFAULT, SWT.DEFAULT);

		// Otherwise an image is stretched by width
		bsize.x = Math.max(bsize.x - 1, bsize.y - 1);
		bsize.y = Math.max(bsize.x - 1, bsize.y - 1);
		button.setSize(bsize);
		shell.setSize(bsize);

		shell.open();

		// Work around a gtk display error that causes the widgets in the
		// shell not being "visible". Opening a second Shell modifies some
		// internal properties of the existing shell, such that the widgets
		// become visible.
		if(getBundle("org.eclipse.swt.gtk.linux.x86_64") != null ||
				getBundle("org.eclipse.swt.gtk.linux.x86") != null) {
			Shell shellGtkQuirk = new Shell(shell, SWT.NO_TRIM);
			shellGtkQuirk.open();
			shellGtkQuirk.close();
		}

		GC gc = new GC(shell);
		Image image = new Image(display, bsize.x, bsize.y);
		gc.copyArea(image, 0, 0);
		gc.dispose();
		shell.close();

		ImageData imageData = image.getImageData();
		imageData.transparentPixel = imageData.palette.getPixel(greenScreen.getRGB());

		image.dispose();
		greenScreen.dispose();

		return new Image(display, imageData);
	}

	/**
	 * Returns an {@link Image} representing the checkbox.
	 * 
	 * @param checked
	 *            Flag if the checkbox is checked.
	 * @param enabled
	 *            Flag if the checkbox is enabled.
	 * 
	 * @return {@link Image} representing the checkbox.
	 */
	public static Image getImage(boolean checked, boolean enabled) {
		if(useTextIcon()) {
			return null;
		}

		ImageRegistry imgReg = getImageRegistry();
		if(imgReg.getDescriptor(CHECKED_KEY) == null) {
			imgReg.put(UNCHECKED_KEY, makeShot(false, true));
			imgReg.put(CHECKED_KEY, makeShot(true, true));
			imgReg.put(DISABLED_UNCHECKED_KEY, makeShot(false, false));
			imgReg.put(DISABLED_CHECKED_KEY, makeShot(true, false));
		}

		if(enabled) {
			return checked ? imgReg.get(CHECKED_KEY) : imgReg.get(UNCHECKED_KEY);
		}

		return checked ? imgReg.get(DISABLED_CHECKED_KEY) : imgReg.get(DISABLED_UNCHECKED_KEY);
	}

	/** {@inheritDoc} */
	@Override
	public Image getImage(Object element) {
		return getImage(isChecked(element), isEnabled(element));
	}

	/** {@inheritDoc} */
	@Override
	public String getText(Object element) {
		if(useTextIcon() && isEnabled(element)) {
			return isChecked(element) ? CHECKED_LABEL : UNCHECKED_LABEL;
		}
		return null;
	}

	/** Tests the model objects checked state. */
	protected abstract boolean isChecked(Object element);

	/** Tests the model objects enabled state. */
	protected abstract boolean isEnabled(Object element);
}
