/*-------------------------------------------------------------------------+
| Copyright 2013 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static java.io.File.separator;
import static java.lang.System.getProperty;
import static java.util.UUID.randomUUID;

/**
 * Utility functions that deal with the underlying system.
 * 
 * @author ratiu, barner
 */
public class SystemUtils {

	/** Cache for {@link #isWindowsPlatform()}. */
	private static Boolean isWindowsPlatform = null;

	/** Cache for {@link #isLinuxPlatform()}. */
	private static Boolean isLinuxPlatform = null;

	/** Cache for {@link #isMacOSXPlatform()}. */
	private static Boolean isMacOSXPlatform = null;

	/** Cache for {@link #isVm64BitArch()}. */
	private static Boolean isVm64BitArch;

	/** Cache for {@link #isOs64BitArch()}. */
	private static Boolean isOs64BitArch;

	/** Predicate if the current platform is Windows. */
	public static boolean isWindowsPlatform() {
		if(isWindowsPlatform != null) {
			return isWindowsPlatform;
		}
		String osName = System.getProperty("os.name");
		isWindowsPlatform = osName != null && osName.contains("Windows");

		return isWindowsPlatform;
	}

	/** Predicate if the current platform is Linux. */
	public static boolean isLinuxPlatform() {
		if(isLinuxPlatform != null) {
			return isLinuxPlatform;
		}
		String osName = System.getProperty("os.name");
		isLinuxPlatform = osName != null && osName.contains("Linux");
		return isLinuxPlatform;
	}

	/** Predicate if the current platform is Mac OS X. */
	public static boolean isMacOSXPlatform() {
		if(isMacOSXPlatform != null) {
			return isMacOSXPlatform;
		}
		String osName = System.getProperty("os.name");
		isMacOSXPlatform = osName != null && osName.contains("Mac OS X");
		return isMacOSXPlatform;
	}

	/** Predicate if the current <b>Java VM</b> architecture is 64 bit. */
	public static boolean isVm64BitArch() {
		if(isVm64BitArch != null) {
			return isVm64BitArch;
		}
		String osArch = System.getProperty("os.arch");
		isVm64BitArch = osArch != null && osArch.contains("64");
		return isVm64BitArch;
	}

	/** Predicate if the current <b>OS</b> architecture is 64 bit. */
	public static boolean isOs64BitArch() {
		if(isOs64BitArch != null) {
			return isOs64BitArch;
		}

		String progFilesX86 = System.getenv("ProgramFiles(x86)");
		isOs64BitArch = isWindowsPlatform() ? progFilesX86 != null : isVm64BitArch();
		return isOs64BitArch;
	}

	/**
	 * Converts an integer value representing an error into a string. If the exit code cannot be
	 * interpreted, then an empty string is returned.
	 */
	public static String interpretProcessExitValues(int exitValue) {
		if(!isWindowsPlatform()) {
			switch(exitValue) {
				case 126:
					return "Command not found";
				case 139:
					return "Segmentation fault";
			}
		}
		return "";
	}

	/** Returns the path to the system's temporary directory. */
	public static String getSystemTempDirPath() {
		return getTempDirPath(false, null);
	}

	/**
	 * Returns the path to a directory in the system's temporary directory that follows the
	 * pattern<br>
	 * {@code <SYS_TEMP_DIR>[/UNIQUE_DIR][/SUB_DIR]}
	 * 
	 * @param unique
	 *            Flag if a unique sub-directory should be appended to the path of the system's
	 *            temporary directory.
	 * @param subdir
	 *            Sub-directory within system's temporary directory (may be {@code null}).
	 * @return Path to a directory in the system's temporary directory.
	 */
	public static String getTempDirPath(boolean unique, String subdir) {
		String tmpDir = getProperty("java.io.tmpdir");
		if(unique) {
			tmpDir += separator + randomUUID();
		}

		if(subdir != null) {
			tmpDir += separator + subdir;
		}

		return tmpDir;
	}
}
