/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures.IndicatorType;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisual;

import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.canvas.Canvas;
import javafx.scene.canvas.GraphicsContext;
import javafx.scene.paint.Color;

/**
 * {@link IVisual} for the diagram background. Note that the {@link Canvas} used here is always as
 * big as the diagram viewer's content area and the grid is painted with an according displacement.
 * Having the canvas as big as the diagram will most certainly result in video memory problems,
 * which requires to implement the background in this special way.
 */
final class GridCanvasVisual implements IVisual {
	/** The diagram bundle. */
	private final DiagramViewer viewer;
	/** The FX canvas to be drawn on. */
	private final Canvas gridCanvas;
	/** The displacement caused by scroll bars. */
	private DiagramCoordinate displacement = new DiagramCoordinate(0, 0);

	/** Constructor. */
	public GridCanvasVisual(DiagramViewer viewer) {
		this.viewer = viewer;
		// the background grid canvas
		this.gridCanvas = new Canvas(100, 100);
		gridCanvas.widthProperty().addListener(evt -> {
			paintGrid();
		});
		gridCanvas.heightProperty().addListener(evt -> {
			paintGrid();
		});
	}

	/** Returns gridCanvas. */
	/* package */Canvas getGridCanvas() {
		return gridCanvas;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableInteraction() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableVisual() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IMVCBundle getMVCBundle() {
		return viewer.getDiagramBundle();
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		if(gridCanvas.getParent() == null) {
			// grid canvas is added below the diagram layers
			viewer.getViewerPane().getChildren().add(0, gridCanvas);
			viewer.getLayers().getMouseState().registerMouseListeners(gridCanvas, getMVCBundle());
			gridCanvas.setOnScroll(viewer.getScrollingHandler());
		}
		paintGrid();
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		viewer.getLayers().getMouseState().unregisterMouseListeners(gridCanvas);
		viewer.getViewerPane().getChildren().remove(gridCanvas);
		gridCanvas.setOnScroll(null);
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getModelBounds() {
		Bounds b = gridCanvas.getBoundsInParent();
		return new Rectangle2D(b.getMinX(), b.getMinY(), b.getWidth(), b.getHeight());
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		DiagramViewerFeatures features = viewer.getFeatures();
		return indication.add(features.getHorizontalSpacing() / 2,
				features.getVerticalSpacing() / 2);
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getCurrentBounds() {
		return getModelBounds();
	}

	/** {@inheritDoc} */
	@Override
	public void requestFocus() {
		gridCanvas.requestFocus();
	}

	/** Paints the grid. */
	public void paintGrid() {
		DiagramViewerFeatures features = viewer.getFeatures();
		double zf = features.getCurrentZoomFactor();
		double width = gridCanvas.getWidth();
		double height = gridCanvas.getHeight();
		GraphicsContext gc = gridCanvas.getGraphicsContext2D();
		// background
		Color backgroundColor = features.getBackgroundColor();
		gc.setFill(backgroundColor);
		gc.setStroke(backgroundColor);
		gc.fillRect(0, 0, width, height);
		// outer border and indicators
		Color indicatorColor = features.getIndicatorColor();
		gc.setStroke(indicatorColor);
		gc.setFill(indicatorColor);
		double hSpacing = features.getHorizontalSpacing();
		double vSpacing = features.getVerticalSpacing();
		double dx = displacement.getX() % 1.0;
		double dy = displacement.getY() % 1.0;
		double zfhs = zf * hSpacing;
		double zfvs = zf * vSpacing;
		if(features.isDrawOuterBorder()) {
			if(displacement.getX() < 1.0) {
				gc.fillRect(0, 0, zfhs * (1.0 - dx), height);
			}
			if(displacement.getY() < 1.0) {
				gc.fillRect(0, 0, width, zfvs * (1.0 - dy));
			}
		}
		// abort if no indicators should be drawn
		IndicatorType indicatorType = features.getIndicatorType();
		if(indicatorType == IndicatorType.INVISIBLE) {
			return;
		}
		// draw grid indicators
		double isize = viewer.getFeatures().getIndicatorSize() * zf;
		for(double x = zfhs * (1.0 - dx); x < width; x += zfhs) {
			for(double y = zfvs * (1.0 - dy); y < height; y += zfvs) {
				drawIndicator(gc, x, y, isize, indicatorType);
			}
		}
	}

	/** Draws the indicators. */
	private void drawIndicator(GraphicsContext gc, double x, double y, double size,
			IndicatorType indicatorType) {
		switch(indicatorType) {
			case CIRCLE:
				drawCircle(gc, x, y, size);
				break;
			case CROSS:
				drawCross(gc, x, y, size);
				break;
			case SQUARE:
				drawSquare(gc, x, y, size);
				break;
			default:
				break;
		}
	}

	/** Draws a small cross at the given point. */
	private void drawCircle(GraphicsContext gc, double x, double y, double size) {
		gc.fillOval(x - size, y - size, 2 * size, 2 * size);
	}

	/** Draws a small cross at the given point. */
	private void drawCross(GraphicsContext gc, double x, double y, double size) {
		gc.strokeLine(x - size, y, x + size, y);
		gc.strokeLine(x, y - size, x, y + size);
	}

	/** Draws a small square at the given point. */
	private void drawSquare(GraphicsContext gc, double x, double y, double size) {
		gc.fillRect(x - size, y - size, 2 * size, 2 * size);
	}

	/** {@inheritDoc} */
	@Override
	public EDragGesture getDragGesture(Node node, DiagramCoordinate diagramLocation) {
		if(node == gridCanvas) {
			return EDragGesture.SELECTION;
		}
		return EDragGesture.NONE;
	}

	/** Sets horizontal scroll displacement. */
	public void setScrollX(double canvasDeltaX) {
		displacement = new DiagramCoordinate(canvasDeltaX, displacement.getY());
		paintGrid();
	}

	/** Sets vertical scroll displacement. */
	public void setScrollY(double canvasDeltaY) {
		displacement = new DiagramCoordinate(displacement.getX(), canvasDeltaY);
		paintGrid();
	}

	/** Sets the width of the grid canvas. */
	public void setWidth(double w) {
		gridCanvas.setWidth(w);
	}

	/** Sets the height of the grid canvas. */
	public void setHeight(double h) {
		gridCanvas.setHeight(h);
	}

	/** Relocates the given scene point to local coordinates. */
	public Point2D sceneToLocal(Point2D locationInScene) {
		return gridCanvas.sceneToLocal(locationInScene);
	}

	/** Returns whether the given node is the grid canvas. */
	public boolean isGridCanvas(Node source) {
		return source == gridCanvas;
	}
}
