/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture.MOVE;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;

/**
 * Default {@link IController} implementation for moving anchorage visuals attached to a parent
 * visual. This controller delegates to the parent controller and tries to acquire an anchorage move
 * controller from it, because in most cases the shape of the parent determines the possible
 * locations of the anchorages.
 */
public class DelegatingContentAnchorageController extends ControllerBase {
	/** Constructor. */
	public DelegatingContentAnchorageController(IContentAnchorageMVCBundle mvcb) {
		super(mvcb);
	}

	/** Returns the {@link IContentAnchorageMVCBundle}. */
	private IContentAnchorageMVCBundle getContentAnchorageBundle() {
		// Wild cast works: see constructor
		return (IContentAnchorageMVCBundle)getMVCBundle();
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(gesture == MOVE && allowMove()) {
			IContentAnchorageMVCBundle contentAnchorageBundle = getContentAnchorageBundle();
			IContentMVCBundle parent = contentAnchorageBundle.getAttachedTo();
			if(parent != null && parent.getController() instanceof AnchorageContentControllerBase) {
				AnchorageContentControllerBase parentController =
						(AnchorageContentControllerBase)parent.getController();
				return parentController.getAnchorageMoveController(contentAnchorageBundle);
			}
			return null;
		}
		return super.getDragController(gesture, node, diagramLocation);
	}

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node,
			DiagramCoordinate lastMouseLocation) {
		return defaultKeyPressControllerWithMove;
	}

	/**
	 * The default implementation for key press controller allowing move with cursor keys and delete
	 * with the delete key.
	 */
	private final IKeyPressController defaultKeyPressControllerWithMove =
			new IKeyPressController() {
				@Override
				public Change keyEvent(KeyEvent event, IMVCBundle bundle, Node source,
						DiagramCoordinate mouseLocation) {
					DiagramViewer viewer = getViewer();
					KeyCode code = event.getCode();
					if(code == KeyCode.DELETE) {
						return new Change() {
							@Override
							public void applyChange() {
								delete();
							}
						};
					}
					if(code == KeyCode.A && event.isControlDown()) {
						viewer.selectAll();
						return null;
					}
					if(allowMove()) {
						if(code == KeyCode.UP || code == KeyCode.DOWN || code == KeyCode.LEFT ||
								code == KeyCode.RIGHT) {
							return keyMove(code);
						}
					}
					return null;
				}
			};

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		if(edg == EDragGesture.MOVE) {
			return allowMove() ? Cursor.MOVE : Cursor.DEFAULT;
		}
		return super.getCurrentCursor(node, diagramLocation);
	}

	/** Returns whether the model element can be moved. */
	protected boolean allowMove() {
		return true;
	}

	/**
	 * Performs the keyboard move by returning a corresponding {@link Change}.
	 * 
	 * @param code
	 *            the arrow key code
	 * @return the {@link Change}, which executes the move operation.
	 */
	protected Change keyMove(KeyCode code) {
		return null;
	}
}
