/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base;

import static java.lang.Math.max;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture.RESIZE_H;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture.RESIZE_V;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange.sizeFeedbackChange;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewer;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.change.Change;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IClickController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.IDragController;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentVisual;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.input.MouseEvent;

/**
 * Base class providing dragging implementation for visuals which move and resize in the
 * {@link DiagramViewer}. Sub-classes must implement {@link #resize(FeedbackChange)} and
 * {@link #move(FeedbackChange)} in order to change the underlying model with the respective
 * coordinate changes.
 */
public abstract class ResizableContentControllerBase extends MoveControllerBase {
	/** Constructor. */
	public ResizableContentControllerBase(IContentMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public IClickController getClickController(Node node, DiagramCoordinate diagramLocation) {
		// wild cast works: see constructor
		IContentVisual visual = (IContentVisual)getVisual();
		if(allowExpandCollapse() && visual.isExpandCollapseWidget(node)) {
			return new ClickControllerBase() {
				/** {@inheritDoc} */
				@Override
				public Change singleClick(MouseEvent event, Node node,
						DiagramCoordinate diagramLocation) {
					return () -> expandCollapse();
				}
			};
		}
		return super.getClickController(node, diagramLocation);
	}

	/** Returns whether the expand/collapse feature is active. */
	public boolean allowExpandCollapse() {
		return false;
	}

	/** Triggers the expand/collapse operation. */
	public void expandCollapse() {
		// the default does nothing
	}

	/** Returns whether the horizontal resize is available. */
	public boolean allowHorizontalResize() {
		return true;
	}

	/** Returns whether the vertical resize is available. */
	public boolean allowVerticalResize() {
		return true;
	}

	/** Returns whether child elements should be resized also. */
	protected boolean resizeChildren() {
		return true;
	}

	/** Sets the current feedback change for the given bundle and possibly its children. */
	private void setResizeFeedbackChange(FeedbackChange change, IMVCBundle bundle) {
		bundle.setFeedbackChange(change);
		if(resizeChildren() && bundle instanceof IContentMVCBundle) {
			IContentMVCBundle cb = (IContentMVCBundle)bundle;
			for(IMVCBundle ch : cb.getChildren()) {
				setResizeFeedbackChange(change, ch);
			}
		}
	}

	/**
	 * The minimum dimension of the diagram element in model coordinates. It will not be resizable
	 * below this value. The default value is a single grid spacing.
	 */
	protected Dimension2D getMinimumSize() {
		DiagramViewerFeatures features = getViewer().getFeatures();
		return new Dimension2D(features.getHorizontalSpacing(), features.getVerticalSpacing());
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node n,
			DiagramCoordinate diagramLocation) {
		if(gesture == RESIZE_H && allowHorizontalResize()) {
			return getHorizontalResizeDragController();
		}
		if(gesture == RESIZE_V && allowVerticalResize()) {
			return getVerticalResizeDragController();
		}
		if(gesture == EDragGesture.RESIZE_VH && allowHorizontalResize() && allowVerticalResize()) {
			return getHorizontalVerticalResizeDragController();
		}
		return super.getDragController(gesture, n, diagramLocation);
	}

	/** Returns the {@link IDragController} for the horizontal resize. */
	public IDragController getHorizontalResizeDragController() {
		return defaultHorizontalResizeDragController;
	}

	/** The default controller for horizontal resize. */
	private IDragController defaultHorizontalResizeDragController = new DragControllerBase() {
		@Override
		public void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			Point2D delta = diagramLocation.subtract(startDiagramLocation);
			// use startBundle since bundle and node might differ
			handleResizeDragInProgress(startBundle, delta.getX(), 0);
		}

		@Override
		public Change safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			Point2D delta = diagramLocation.subtract(startDiagramLocation);
			// use startBundle since bundle and node might differ
			return handleResizeCompleted(startBundle, delta.getX(), 0);
		}
	};

	/** Applies the feedback change for the given delta. */
	private void handleResizeDragInProgress(IMVCBundle bundle, double dw, double dh) {
		FeedbackChange altered = alterResizeFeedback(sizeFeedbackChange(dw, dh));
		setResizeFeedbackChange(altered, bundle);
	}

	/** Handles the completion of the resize operation. */
	private Change handleResizeCompleted(IMVCBundle bundle, double dw, double dh) {
		bundle.setFeedbackChange(FeedbackChange.EMPTY);
		final FeedbackChange altered = alterResizeFeedback(sizeFeedbackChange(dw, dh));
		if(altered.isEmpty()) {
			return null;
		}
		return () -> resize(altered);
	}

	/**
	 * Returns whether resize operations should be clamped to positive diagram area and snapped to
	 * the grid.
	 */
	protected boolean clampAndSnapResize() {
		return true;
	}

	/** Alters the given resize feedback (e.g. to uphold minimum size. */
	protected FeedbackChange alterResizeFeedback(FeedbackChange change) {
		IContentVisual visual = (IContentVisual)getMVCBundle().getVisual();
		Rectangle2D orig = visual.getModelBounds();
		double dw = change.getDeltaW();
		double dh = change.getDeltaH();
		Dimension2D minDim = getMinimumSize();
		double newMaxX = orig.getMinX() + max(minDim.getWidth(), orig.getWidth() + dw);
		double newMaxY = orig.getMinY() + max(minDim.getHeight(), orig.getHeight() + dh);
		if(clampAndSnapResize()) {
			DiagramViewerFeatures features = getViewer().getFeatures();
			if(features.useSnapToGrid()) {
				newMaxX = features.snapToGridX(newMaxX);
				newMaxY = features.snapToGridY(newMaxY);
			}
		}
		return sizeFeedbackChange(newMaxX - orig.getMaxX(), newMaxY - orig.getMaxY());
	}

	/** Returns the {@link IDragController} for the vertical resize. */
	public IDragController getVerticalResizeDragController() {
		return defaultVerticalResizeController;
	}

	/** The default controller for vertical resize. */
	private IDragController defaultVerticalResizeController = new DragControllerBase() {
		@Override
		public void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			Point2D delta = diagramLocation.subtract(startDiagramLocation);
			// use startBundle since bundle and node might differ
			handleResizeDragInProgress(startBundle, 0, delta.getY());
		}

		@Override
		public Change safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			Point2D delta = diagramLocation.subtract(startDiagramLocation);
			// use startBundle since bundle and node might differ
			return handleResizeCompleted(startBundle, 0, delta.getY());
		}
	};

	/** Returns the {@link IDragController} for the resize. */
	public IDragController getHorizontalVerticalResizeDragController() {
		return defaultHorizontalVerticalResizeDragController;
	}

	/** The default controller for horizontal and vertical resize. */
	private IDragController defaultHorizontalVerticalResizeDragController =
			new DragControllerBase() {
				@Override
				public void safeDragInProgress(IMVCBundle bundle, Node node,
						DiagramCoordinate diagramLocation) {
					Point2D delta = diagramLocation.subtract(startDiagramLocation);
					// use startBundle since bundle and node might differ
					handleResizeDragInProgress(startBundle, delta.getX(), delta.getY());
				}

				@Override
				public Change safeDragCompleted(IMVCBundle bundle, Node node,
						DiagramCoordinate diagramLocation) {
					Point2D delta = diagramLocation.subtract(startDiagramLocation);
					// use startBundle since bundle and node might differ
					return handleResizeCompleted(startBundle, delta.getX(), delta.getY());
				}
			};

	/**
	 * Resizes the model element.
	 * 
	 * @param delta
	 *            the delta width and height to be applied to the model element
	 */
	protected void resize(FeedbackChange delta) {
		// default does nothing
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		switch(edg) {
			case RESIZE_H:
				return allowHorizontalResize() ? Cursor.H_RESIZE : Cursor.DEFAULT;
			case RESIZE_V:
				return allowVerticalResize() ? Cursor.V_RESIZE : Cursor.DEFAULT;
			case RESIZE_VH:
				return allowHorizontalResize() && allowVerticalResize() ? Cursor.SE_RESIZE
						: Cursor.DEFAULT;
			default:
		}
		return super.getCurrentCursor(node, diagramLocation);
	}
}
