/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.controller.elliptic;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.fortiss.tooling.common.ui.javafx.lwfxef.FeedbackChange;
import org.fortiss.tooling.common.ui.javafx.lwfxef.controller.base.LinkControllerBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.ILinkVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic.CurveLinkVisualBase;

import javafx.geometry.Point2D;

/** Base class for {@link CurveLinkVisualBase} dragging implementation. */
public abstract class CurveLinkBendPointControllerBase extends LinkControllerBase {
	/** Constructor. */
	public CurveLinkBendPointControllerBase(ILinkMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	protected final void moveBendPointVisually(int bpIndex, Point2D delta) {
		super.moveBendPointVisually(bpIndex, delta);
		if(bpIndex <= 0 || bpIndex >= getNumerOfBendPoints() - 1) {
			// first and last bend-point move unconnected
			return;
		}
		ILinkVisual linkVisual = getLinkVisual();
		FeedbackChange centerPointChange = getMoveDeltaFeedback(bpIndex, delta);
		int indexMod3 = bpIndex % 3;
		if(indexMod3 == 2) {
			// bend-points between curve segments also drag the previous and next point
			linkVisual.setFeedbackChangeForBendPoint(bpIndex - 1, centerPointChange);
			linkVisual.setFeedbackChangeForBendPoint(bpIndex + 1, centerPointChange);
		} else {
			// otherwise the connected bend-point is moved in the opposite direction
			FeedbackChange inv = locationFeedbackChange(-centerPointChange.getDeltaX(),
					-centerPointChange.getDeltaY());
			int otherIndex = (indexMod3 == 1) ? bpIndex + 2 : bpIndex - 2;
			linkVisual.setFeedbackChangeForBendPoint(otherIndex, inv);
		}
		// linkVisual.update();
		return;
	}

	/** {@inheritDoc} */
	@Override
	protected final void moveBendPointInModel(int bpIndex, Point2D delta) {
		FeedbackChange chg = getMoveDeltaFeedback(bpIndex, delta);
		getLinkVisual().setFeedbackChangeForBendPoint(bpIndex, null);
		updateModelAfterBendPointMove(bpIndex, chg.getDeltaX(), chg.getDeltaY());
		if(bpIndex <= 0 || bpIndex >= getNumerOfBendPoints() - 1) {
			// first and last bend-point move unconnected
			return;
		}
		ILinkVisual linkVisual = getLinkVisual();
		int indexMod3 = bpIndex % 3;
		if(indexMod3 == 2) {
			// bend-points between curve segments also drag the previous and next point
			linkVisual.setFeedbackChangeForBendPoint(bpIndex - 1, null);
			updateModelAfterBendPointMove(bpIndex - 1, chg.getDeltaX(), chg.getDeltaY());
			linkVisual.setFeedbackChangeForBendPoint(bpIndex + 1, null);
			updateModelAfterBendPointMove(bpIndex + 1, chg.getDeltaX(), chg.getDeltaY());
		} else {
			// otherwise the connected bend-point is moved in the opposite direction
			int otherIndex = (indexMod3 == 1) ? bpIndex + 2 : bpIndex - 2;
			linkVisual.setFeedbackChangeForBendPoint(otherIndex, null);
			updateModelAfterBendPointMove(otherIndex, -chg.getDeltaX(), -chg.getDeltaY());
		}
	}

	/** {@inheritDoc} */
	@Override
	protected final void deleteBendPoint(int bpIndex) {
		if(bpIndex <= 0 || bpIndex >= getNumerOfBendPoints() - 1) {
			// never delete first and last bend-points (i.e. control points)
			return;
		}
		int indexMod3 = bpIndex % 3;
		if(indexMod3 == 0) {
			deleteFromCurve(bpIndex - 2, bpIndex - 1, bpIndex);
		} else if(indexMod3 == 1) {
			deleteFromCurve(bpIndex, bpIndex + 1, bpIndex + 2);
		} else {
			deleteFromCurve(bpIndex - 1, bpIndex, bpIndex + 1);
		}
	}

	/** Deletes the given three bend-points from the model. */
	protected abstract void deleteFromCurve(int ctrlPoint2Index, int segmentPointIndex,
			int ctrlPoint1Index);

	/** {@inheritDoc} */
	@Override
	protected final void createBendPointAt(int bpIndex, DiagramCoordinate location) {
		DiagramViewerFeatures features = getViewer().getFeatures();
		DiagramCoordinate c2 = new DiagramCoordinate(
				location.getX() - 2 * features.getHorizontalSpacing(), location.getY());
		DiagramCoordinate c1 = new DiagramCoordinate(
				location.getX() + 2 * features.getHorizontalSpacing(), location.getY());
		createCurveSegment(bpIndex - 1, c2, location, c1);
	}

	/** Creates the new curve segment with the given locations after the given index. */
	protected abstract void createCurveSegment(int afterIndex, DiagramCoordinate newCtrlPt2,
			DiagramCoordinate newSegmentPoint, DiagramCoordinate newCtrlPt1);
}
