/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.FOCUS_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HOVER_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_ALLOWED_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_DENIED_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.LINK_TARGET_MAYBE_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.PRIMARY_SELECTION_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.SECONDARY_SELECTION_TAG;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.impl.MVCBundlePartBase;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IVisual;

/** {@link MVCBundlePartBase} base class extended with effect methods from {@link IVisual}. */
public abstract class MVCBundlePartWithEffectsBase extends MVCBundlePartBase implements IVisual {
	/** Constructor. */
	public MVCBundlePartWithEffectsBase(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		IMVCBundle bundle = getMVCBundle();
		// selection and focus effects
		if(bundle.hasTag(PRIMARY_SELECTION_TAG)) {
			if(bundle.hasTag(FOCUS_TAG)) {
				applySelectedFocusedEffect(layers);
			} else {
				applySelectedNotFocusedEffect(layers);
			}
		} else if(bundle.hasTag(SECONDARY_SELECTION_TAG)) {
			applySecondarySelectedEffect(layers);
		} else {
			applyNotSelectedEffect(layers);
		}
		// hover effects
		if(bundle.hasTag(HOVER_TAG)) {
			createHoverEffect(layers);
		} else {
			removeHoverEffect(layers);
		}
		// link target effects
		if(bundle.hasTag(LINK_TARGET_ALLOWED_TAG)) {
			createLinkTargetAllowedEffect(layers);
		} else if(bundle.hasTag(LINK_TARGET_DENIED_TAG)) {
			createLinkTargetDeniedEffect(layers);
		} else if(bundle.hasTag(LINK_TARGET_MAYBE_TAG)) {
			createLinkTargetMaybeEffect(layers);
		} else {
			createLinkTargetDisabledEffect(layers);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		applyNotSelectedEffect(layers);
		removeHoverEffect(layers);
		createLinkTargetDisabledEffect(layers);
	}

	/**
	 * Applies the effect when not selected and not focused.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void applyNotSelectedEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Applies the effect when selected and focused.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void applySelectedFocusedEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Applies the effect when selected, but not focused.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void applySelectedNotFocusedEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Applies the effect when secondary selected.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void applySecondarySelectedEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Creates the hover effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void createHoverEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Removes the hover effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void removeHoverEffect(DiagramLayers layers) {
		// the default does nothing
	}

	/**
	 * Creates the link target allowed effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void createLinkTargetAllowedEffect(DiagramLayers layers) {
		// do nothing
	}

	/**
	 * Creates the link target denied effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void createLinkTargetDeniedEffect(DiagramLayers layers) {
		// do nothing
	}

	/**
	 * Creates the link target maybe effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void createLinkTargetMaybeEffect(DiagramLayers layers) {
		// do nothing
	}

	/**
	 * Creates the link target disabled effect.
	 * 
	 * @param layers
	 *            the layers of the diagram
	 */
	protected void createLinkTargetDisabledEffect(DiagramLayers layers) {
		// do nothing
	}
}
