/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic;

import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_INCOMING_LINK_TAG;
import static org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_OUTGOING_LINK_TAG;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramLayers;
import org.fortiss.tooling.common.ui.javafx.lwfxef.EDragGesture;
import org.fortiss.tooling.common.ui.javafx.lwfxef.mvc.IContentAnchorageMVCBundle;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.IContentAnchorageVisual;
import org.fortiss.tooling.common.ui.javafx.lwfxef.visual.base.ContentAnchorageVisualBase;

import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;

/** Base class for {@link IContentAnchorageVisual}s depicted as circles. */
public abstract class CircularContentAnchorageVisualBase extends ContentAnchorageVisualBase {
	/** The default insets. */
	private static final double DEFAULT_INSET = 2;

	/** Constructor. */
	public CircularContentAnchorageVisualBase(IContentAnchorageMVCBundle mvcb) {
		super(mvcb, new Circle(), new Circle());
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);
		updateCircleProperties();
	}

	/** Sets all properties of the main node. */
	private void updateCircleProperties() {
		Rectangle2D b = getCurrentBounds();
		if(enableVisual()) {
			Circle circle = getVisualCircle();
			circle.setCenterX(b.getMinX() + b.getWidth() / 2);
			circle.setCenterY(b.getMinY() + b.getHeight() / 2);
			circle.setRadius(getInnerRadius(b));
			circle.setOpacity(getOpacity());
			circle.setStrokeWidth(getBorderWidth());
			circle.setStrokeType(getBorderType());
			if(getMVCBundle().hasTag(HIGHLIGHT_INCOMING_LINK_TAG)) {
				circle.setFill(getHighlightIncomingLinkColor());
				circle.setStroke(getHighlightIncomingLinkBorderColor());
			} else if(getMVCBundle().hasTag(HIGHLIGHT_OUTGOING_LINK_TAG)) {
				circle.setFill(getHighlightOutgoingLinkColor());
				circle.setStroke(getHighlightOutgoingLinkBorderColor());
			} else {
				circle.setFill(getFillColor());
				circle.setStroke(getBorderColor());
			}
		}

		if(enableInteraction()) {
			Circle ha = getHitAreaCircle();
			ha.setCenterX(b.getMinX() + b.getWidth() / 2);
			ha.setCenterY(b.getMinY() + b.getHeight() / 2);
			ha.setRadius(getOuterRadius(b));
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** Helper function to compute outer radius. */
	private double getOuterRadius(Rectangle2D b) {
		return getInnerRadius(b) + getHitAreaOutset();
	}

	/** Helper function to compute inner radius. */
	private double getInnerRadius(Rectangle2D b) {
		return b.getWidth() / 2 - getInset();
	}

	/** Returns the insets of the filled circle subtracted from {@link #getDimensions()}. */
	protected double getInset() {
		return DEFAULT_INSET;
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node n, DiagramCoordinate diagramLocation) {
		if(n == getHitAreaCircle()) {
			Circle ha = getHitAreaCircle();
			double dx = Math.abs(diagramLocation.getX() - ha.getCenterX());
			double dy = Math.abs(diagramLocation.getY() - ha.getCenterY());
			Rectangle2D bounds = getCurrentBounds();
			double innerRadius = getInnerRadius(bounds);
			double outerRadius = getOuterRadius(bounds);
			double squaredLocation = dx * dx + dy * dy;
			if(squaredLocation > innerRadius * innerRadius &&
					squaredLocation <= outerRadius * outerRadius) {
				return EDragGesture.NEW_LINK;
			}
			return EDragGesture.MOVE;
		}
		if(n == getVisualShape()) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}

	/** Returns the visual circle. */
	protected final Circle getVisualCircle() {
		// wild cast works: see constructor
		return (Circle)getVisualShape();
	}

	/** Returns the hit area circle. */
	protected final Circle getHitAreaCircle() {
		// wild cast works: see constructor
		return (Circle)getHitAreaShape();
	}
}
