/*******************************************************************************
 * Copyright (c) 2019, 2020 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0 which is available at
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.lwfxef.visual.elliptic;

import static java.lang.Math.atan2;
import static java.util.Objects.requireNonNull;

import org.fortiss.tooling.common.ui.javafx.lwfxef.DiagramCoordinate;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;

/** This class encapsulates locations on the border of an ellipse defined by the angular value. */
public final class EllipticBorderLocation {
	/** The angle in degrees. */
	private final double angleInDegree;
	/** The radiuses of the ellipse. */
	private final Dimension2D radiuses;
	/**
	 * The correction values used to compute the final location (usually half the size of the object
	 * located on the border so that it looks centered.
	 */
	private final Dimension2D correction;

	/** Constructor. */
	public EllipticBorderLocation(double angleInDegree, Dimension2D radiuses,
			Dimension2D correction) {
		this.angleInDegree = angleInDegree;
		this.radiuses = requireNonNull(radiuses);
		this.correction = requireNonNull(correction);
	}

	/** Returns the angle in degree. */
	public double getAngleInDegree() {
		return angleInDegree;
	}

	/** Returns the radiuses. */
	public Dimension2D getRadiuses() {
		return radiuses;
	}

	/**
	 * Returns the location computed from this {@link EllipticBorderLocation} relative to the
	 * ellipse center.
	 * TODO(JE): leads to java.lang.ArithmeticException: / by zero if cos is 0. What is the
	 * difference to the existing code? Why is an adaption needed?
	 */
	public DiagramCoordinate getLocation() {
		double angle = Math.PI * angleInDegree / 180;
		double rh = radiuses.getWidth();
		double rv = radiuses.getHeight();
		DiagramCoordinate middle = new DiagramCoordinate(rh, rv);
		double cos = Math.cos(angle);
		double sin = Math.sin(angle);
		double tan = (sin / cos);
		// Division by zero will result in tan being Infinity,
		// which gives the right results for x and y.
		double x = rh * rv / Math.sqrt(rv * rv + rh * rh * tan * tan);
		double y = rh * rv * Math.signum(tan) / Math.sqrt(rv * rv / (tan * tan) + rh * rh);
		return middle.add(x * Math.signum(cos), -y * Math.signum(cos))
				.subtract(correction.getWidth() / 2.0, correction.getHeight() / 2.0);
	}

	/**
	 * Returns the closest point on the ellipse bounds inside the given rectangle when pointing at
	 * the given location.
	 */
	public static final EllipticBorderLocation getClosestLocationOnBounds(Point2D point,
			Rectangle2D ellipse, Dimension2D size) {
		double w2 = ellipse.getWidth() / 2;
		double h2 = ellipse.getHeight() / 2;
		double dx = point.getX() - w2;
		double dy = h2 - point.getY();

		double angle = atan2(dy, dx) * 180.0 / Math.PI;
		Dimension2D radiuses = new Dimension2D(w2, h2);
		return new EllipticBorderLocation(angle, radiuses, size);
	}
}
