/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.internal.views;

import static java.util.Arrays.asList;
import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.service.IPrototypeService.PROTOTYPE_DATA_FORMAT;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.filterOutInstanceOf;
import static org.fortiss.tooling.kernel.utils.KernelModelElementUtils.getParentElement;

import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.function.Predicate;

import org.conqat.ide.commons.ui.ui.EmptyPartListener;
import org.conqat.lib.commons.collections.IdentityHashSet;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.fx.ui.workbench3.FXViewPart;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeContentProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeUIProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeViewer;
import org.fortiss.tooling.kernel.extension.data.LibraryPrototype;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.extension.data.PrototypeCategory;
import org.fortiss.tooling.kernel.model.ILibraryElement;
import org.fortiss.tooling.kernel.service.IPrototypeService;
import org.fortiss.tooling.kernel.ui.extension.base.EditorBase;
import org.fortiss.tooling.kernel.ui.internal.editor.ExtendableMultiPageEditor;
import org.fortiss.tooling.kernel.ui.listener.ExtendableMultiPageEditorPageChangeListener;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;

import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.TextField;
import javafx.scene.input.ClipboardContent;
import javafx.scene.layout.BorderPane;

/**
 * {@link FXViewPart} for showing composable model elements with DND editing support.
 * 
 * @author hoelzl
 */
public final class ModelElementsViewFX extends FXViewPart {
	/** A dummy object serving as hidden root element of the tree view. */
	private static final Object DUMMY_ROOT = new Object();
	/** The tree-table viewer. */
	private DynamicTreeViewer<Object> treeViewer;
	/** The filter text field if filtering is enabled in content provider. */
	private TextField filterWidget;
	/** Current active {@link ExtendableMultiPageEditor}. */
	private ExtendableMultiPageEditor activeBindingEditor = null;
	/** Stores the editor activation listener. */
	private EditorActivationListener editorActivationListener = new EditorActivationListener();
	/** The container object used. */
	private EObject containerObject = null;
	/** The prototypes supported by the current editor. */
	private Set<Prototype> supportedBaseClasses = new IdentityHashSet<Prototype>();
	/** Listener for reacting to changes of the embedded editor. */
	private final ExtendableMultiPageEditorPageChangeListener bindingEditorPageChangeListener =
			new ExtendableMultiPageEditorPageChangeListener() {
				@Override
				public void pageChanged() {
					updateEditorFilters(activeBindingEditor.getActiveModelEditor());
				}
			};

	/** {@inheritDoc} */
	@Override
	protected Scene createFxScene() {
		getSite().getWorkbenchWindow().getPartService().addPartListener(editorActivationListener);

		BorderPane pane = new BorderPane();
		ContentProvider contentProvider = new ContentProvider();
		UIProvider uiProvider = new UIProvider();

		filterWidget = new TextField();
		filterWidget.textProperty().addListener((obs, oVal, nVal) -> {
			contentProvider.setFilterExpression(nVal);
			treeViewer.update();
		});

		treeViewer = new DynamicTreeViewer<>(DUMMY_ROOT, false, 2, contentProvider, uiProvider);

		pane.setTop(filterWidget);
		pane.setCenter(treeViewer.getControl());

		return new Scene(pane);
	}

	/** {@inheritDoc} */
	@Override
	protected void setFxFocus() {
		filterWidget.requestFocus();
	}

	/** Updates the filters according to the given editor. */
	@SuppressWarnings("unchecked")
	private void updateEditorFilters(IEditorPart editor) {
		supportedBaseClasses.clear();
		if(editor instanceof EditorBase &&
				((EditorBase<? extends EObject>)editor).enableLibraryView()) {
			EditorBase<? extends EObject> editorBase = (EditorBase<? extends EObject>)editor;
			containerObject = editorBase.getEditedObject();

			for(Class<? extends EObject> clazz : editorBase.getVisibleEObjectTypes()) {
				List<Prototype> composablePrototypes =
						IPrototypeService.getInstance().getComposablePrototypes(clazz);
				if(getParentElement(containerObject, ILibraryElement.class, true) == null) {
					supportedBaseClasses.addAll(composablePrototypes);
				} else {
					supportedBaseClasses.addAll(
							filterOutInstanceOf(LibraryPrototype.class, composablePrototypes));
				}
			}
		} else {
			containerObject = null;
		}

		treeViewer.update();
		treeViewer.expandAllItems();
	}

	/** Switches to the given workbench editor part. */
	private void switchWorkbenchEditor(IEditorPart editor) {
		if(activeBindingEditor != null) {
			activeBindingEditor.removeBindingEditorListener(bindingEditorPageChangeListener);
			activeBindingEditor = null;
		}

		if(editor instanceof ExtendableMultiPageEditor) {
			activeBindingEditor = (ExtendableMultiPageEditor)editor;
			activeBindingEditor.addBindingEditorListener(bindingEditorPageChangeListener);
			updateEditorFilters(activeBindingEditor.getActiveModelEditor());
		} else if(editor != null) {
			updateEditorFilters(editor);
		}
	}

	/** {@link DynamicTreeContentProviderBase} for the library view. */
	private class ContentProvider extends DynamicTreeContentProviderBase<Object> {
		/** {@inheritDoc} */
		@Override
		protected Collection<? extends Object> getChildren(Object parent) {
			if(parent instanceof PrototypeCategory) {
				PrototypeCategory protoCat = (PrototypeCategory)parent;
				return asList(protoCat.getChildren());
			}
			if(parent == DUMMY_ROOT) {
				return IPrototypeService.getInstance().getAllTopLevelPrototypesCategories();
			}
			return emptyList();
		}

		/** {@inheritDoc} */
		@Override
		protected boolean filter(Object element, String filterValue) {
			if(containerObject == null || supportedBaseClasses.isEmpty()) {
				return false;
			}
			if(element instanceof PrototypeCategory) {
				PrototypeCategory protoCat = (PrototypeCategory)element;
				for(Object child : protoCat.getChildren()) {
					if(filter(child, filterValue)) {
						return true;
					}
				}
				return false;
			}
			if(supportedBaseClasses.contains(element)) {
				if(filterValue == null || "".equals(filterValue)) {
					return true;
				}
				if(element instanceof Prototype) {
					Prototype proto = (Prototype)element;
					return proto.getName().toLowerCase().contains(filterValue.toLowerCase());
				}
			}
			return false;
		}

		/** {@inheritDoc} */
		@Override
		protected Predicate<Object> getFilterPredicate(String filterValue) {
			// this override is needed, because super implementation does not
			// call filter(Object, String) when filter value is null.
			if(filterValue == null || "".equals(filterValue)) {
				return (o) -> {
					return filter(o, null);
				};
			}
			return super.getFilterPredicate(filterValue);
		}
	}

	/** {@link DynamicTreeUIProviderBase} for the library view. */
	private class UIProvider extends DynamicTreeUIProviderBase<Object> {
		/** {@inheritDoc} */
		@Override
		public String getLabel(Object element) {
			if(element instanceof Prototype) {
				Prototype proto = (Prototype)element;
				return proto.getName();
			}
			if(element instanceof PrototypeCategory) {
				PrototypeCategory protoCat = (PrototypeCategory)element;
				return protoCat.getName();
			}
			return element == null ? "" : element.toString();
		}

		/** {@inheritDoc} */
		@Override
		public Node getIconNode(Object element) {
			if(element instanceof Prototype) {
				Prototype proto = (Prototype)element;
				EObject eoPrototype = proto.getPrototype();
				return IModelElementHandlerService.getInstance().getFXIcon(eoPrototype);
			}
			if(element instanceof PrototypeCategory) {
				PrototypeCategory protoCat = (PrototypeCategory)element;
				if(protoCat.getChildren().length > 0) {
					return getIconNode(protoCat.getChildren()[0]);
				}
			}
			return null;
		}

		/** {@inheritDoc} */
		@Override
		public ClipboardContent getDragClipboardContent(Object element) {
			if(element instanceof Prototype) {
				Prototype proto = (Prototype)element;
				String uid = IPrototypeService.getInstance().getUniquePrototypeID(proto);
				if(uid != null) {
					ClipboardContent cbContent = new ClipboardContent();
					cbContent.put(PROTOTYPE_DATA_FORMAT, uid);
					return cbContent;
				}
			}
			return null;
		}
	}

	/** If an editor in a different Project is opened the Model is reinitialized */
	private class EditorActivationListener extends EmptyPartListener {
		/** Change the tree viewer content whenever workbench part changes. */
		@Override
		public void partActivated(IWorkbenchPart workbenchPart) {
			if(!(workbenchPart instanceof IEditorPart)) {
				return;
			}

			IEditorPart part = (IEditorPart)workbenchPart;
			treeViewer.update();
			switchWorkbenchEditor(part);
		}
	}
}
