/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.control;

import static org.fortiss.tooling.spiderchart.util.AxisUtils.getSegmentPoint;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.spiderchart.model.AxisBase;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.DoubleAxis;
import org.fortiss.tooling.spiderchart.model.EnumerationAxis;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.AxisStyle;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.DataSeriesStyle;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.scene.canvas.GraphicsContext;

/**
 * Class for drawing the spider chart.
 * 
 * @author hoelzl
 */
public final class SpiderChartControl extends SpiderChartControlBase {
	/** The center point of the chart. */
	private Point2D center = new Point2D(0, 0);
	/** The end points of the axes. */
	private Map<AxisBase, Point2D> axesEnds = new HashMap<>();
	/** The angle in degree between two axes. */
	private double segmentDegree = 0.0;

	/** Constructor. */
	public SpiderChartControl(SpiderChart chart, ChartStyle style) {
		super(chart, style);
	}

	/** Draws the spider chart. */
	public void draw(GraphicsContext gc) {
		if(!style.isUseIndividualAxisSegments()) {
			drawBackground(gc);
			drawWebLines(gc);
		}
		drawAxes(gc);
		for(DataSeries data : chart.getDataSeries()) {
			drawDataSeries(gc, data);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void setBounds(double x, double y, double w, double h) {
		super.setBounds(x, y, w, h);
		computeAxisEndPoints();
	}

	/** Computes the end points of the axes. */
	private void computeAxisEndPoints() {
		// compute the axis points
		List<AxisBase> axes = chart.getAxes();
		int size = axes.size();
		segmentDegree = 360.0 / size;
		double currentAngleDegree = style.getStartAngleDegree();

		double halfWidth = getWidth() / 2;
		double halfHeight = getHeight() / 2;
		center = new Point2D(getX() + halfWidth, getY() + halfHeight);

		double relativeMargin = 1.0 - style.getRelativeMargin();
		int wExtent = (int)(relativeMargin * halfWidth);
		int hExtent = (int)(relativeMargin * halfHeight);
		int extent = Math.min(wExtent, hExtent);

		for(int i = 0; i < size; i++) {
			AxisBase axis = axes.get(i);
			// compute end point
			double startAngleRad = Math.PI * currentAngleDegree / 180.0;
			int extentX = (int)(Math.cos(startAngleRad) * extent);
			int extentY = (int)(Math.sin(startAngleRad) * extent);
			Point2D outer = new Point2D(center.getX() + extentX, center.getY() + extentY);
			axesEnds.put(axis, outer);
			currentAngleDegree = (currentAngleDegree + segmentDegree) % 360.0;
		}
	}

	/** Draws the background if style requires all axes to have equal number of segments. */
	private void drawBackground(GraphicsContext gc) {
		gc.save();
		if(chart.getAxes().isEmpty()) {
			return;
		}

		gc.beginPath();
		boolean first = true;
		for(AxisBase axis : chart.getAxes()) {
			Point2D p = axesEnds.get(axis);
			if(first) {
				gc.moveTo(p.getX(), p.getY());
				first = false;
			} else {
				gc.lineTo(p.getX(), p.getY());
			}
		}
		gc.closePath();
		style.getBackgroundFillStyle().applyStyle(gc);
		gc.fill();
		gc.restore();
	}

	/** Draws the web lines of the background if all axes have equal number of segments. */
	private void drawWebLines(GraphicsContext gc) {
		style.getBackgroundLineStyle().applyStyle(gc);
		for(int i = 1; i <= style.getAxisSegments(); i++) {
			computeAndDrawPath(i, gc);
		}
	}

	/** Computes the i-th web line path. */
	private void computeAndDrawPath(int i, GraphicsContext gc) {
		double ratio = (double)i / (double)style.getAxisSegments();
		boolean first = true;
		gc.save();
		gc.beginPath();
		for(AxisBase axis : chart.getAxes()) {
			Point2D outer = axesEnds.get(axis);
			double xExtent = outer.getX() - center.getX();
			double yExtent = outer.getY() - center.getY();
			double x = center.getX() + ratio * xExtent;
			double y = center.getY() + ratio * yExtent;
			if(first) {
				gc.moveTo(x, y);
				first = false;
			} else {
				gc.lineTo(x, y);
			}
		}
		gc.closePath();
		gc.restore();
	}

	/** Draws the data series. */
	private void drawDataSeries(GraphicsContext gc, DataSeries data) {
		DataSeriesStyle dStyle = style.getDataSeriesStyle(data);
		boolean first = true;
		gc.save();
		gc.beginPath();
		for(AxisBase axis : chart.getAxes()) {
			Point2D p = computeDataPoint(axis, data);
			if(first) {
				gc.moveTo(p.getX(), p.getY());
				first = false;
			} else {
				gc.lineTo(p.getX(), p.getY());
			}
			if(dStyle.isShowIndicators()) {
				drawIndicator(gc, dStyle, p);
			}
			if(dStyle.isShowIndicatorLabels()) {
				drawIndicatorLabel(gc, dStyle, p, axis, data.getAxisValue(axis));
			}
		}
		gc.closePath();
		dStyle.getFillStyle().applyStyle(gc);
		gc.fill();
		dStyle.getLineStyle().applyStyle(gc);
		gc.stroke();
		gc.restore();
	}

	/** Draws the indicator labels next to the chart point. */
	private void drawIndicatorLabel(GraphicsContext gc, DataSeriesStyle dStyle, Point2D p,
			AxisBase axis, Object value) {
		FontStyle fontStyle = dStyle.getIndicatorLabelStyle();
		String label = dStyle.getFormattedLabel(axis, value);
		Dimension2D textExtent = fontStyle.getTextBounds(label);
		// data series indicator labels are always to the right of the point
		int offset = 5;
		double x = p.getX() + offset;
		double y = p.getY();
		if(y > center.getY()) {
			y = y + offset + textExtent.getHeight() / 2;
		} else if(y < center.getY()) {
			y = y - offset;
		} else {
			y = y - textExtent.getHeight() / 2;
		}
		fontStyle.drawText(gc, label, x, y);
	}

	/** Draws the filled indicator rectangle. */
	private void drawIndicator(GraphicsContext gc, DataSeriesStyle dStyle, Point2D p) {
		double size = dStyle.getIndicatorSize();
		dStyle.getLineStyle().applyStyle(gc);
		gc.fillRect(p.getX() - size / 2, p.getY() - size / 2, size, size);
	}

	/** Computes the point on the given axis for the given data. */
	@SuppressWarnings("unchecked")
	private Point2D computeDataPoint(AxisBase axis, DataSeries data) {
		Double ratio = 0.0;
		if(axis instanceof DoubleAxis) {
			DoubleAxis da = (DoubleAxis)axis;
			ratio = da.getAxisRatio((Double)data.getAxisValue(axis));
		}
		if(axis instanceof EnumerationAxis) {
			EnumerationAxis<Object> ea = (EnumerationAxis<Object>)axis;
			int segments = style.isUseIndividualAxisSegments()
					? style.getAxisStyle(axis).getSegments() : style.getAxisSegments();
			int index = Math.max(0, ea.indexOfEnumerationMember(data.getAxisValue(ea)));
			ratio = (double)index / (double)segments;
		}
		Point2D outer = axesEnds.get(axis);
		int deltaX = (int)(ratio * (outer.getX() - center.getX()));
		int deltaY = (int)(ratio * (outer.getY() - center.getY()));
		return new Point2D(center.getX() + deltaX, center.getY() + deltaY);
	}

	/** Draws the axes of the spider chart. */
	private Map<AxisBase, Point2D> drawAxes(GraphicsContext gc) {
		List<AxisBase> axes = chart.getAxes();
		int size = axes.size();
		double segmentDegree = 360.0 / size;
		double angleDegree = 90.0;

		double halfWidth = getWidth() / 2;
		double halfHeight = getHeight() / 2;
		double centerX = getX() + halfWidth;
		double centerY = getY() + halfHeight;

		for(int i = 0; i < size; i++) {
			AxisBase axis = axes.get(i);
			Point2D outer = axesEnds.get(axis);
			// draw line
			AxisStyle aStyle = style.getAxisStyle(axis);
			aStyle.getLineStyle().drawLine(gc, centerX, centerY, outer.getX(), outer.getY());
			// draw label
			drawAxisLabel(gc, axis, aStyle, centerX, centerY, outer.getX(), outer.getY());

			if(aStyle.isUseSegmentIndicators()) {
				drawSegmentIndicators(gc, centerX, centerY, axis, outer, aStyle);
			}
			angleDegree = (angleDegree + segmentDegree) % 360.0;
		}
		return axesEnds;
	}

	/** Draws the indicators of the given axis. */
	private void drawSegmentIndicators(GraphicsContext gc, double centerX, double centerY,
			AxisBase axis, Point2D outer, AxisStyle aStyle) {
		// draw segment indicators
		int segments = style.isUseIndividualAxisSegments() ? aStyle.getSegments()
				: style.getAxisSegments();
		for(int s = 1; s <= segments; s++) {
			String lbl = "";
			if(axis instanceof DoubleAxis) {
				double ratio = (double)s / (double)segments;
				Double value = ((DoubleAxis)axis).getAxisValue(ratio);
				lbl = aStyle.getDecimalFormat().format(value);
			} else if(axis instanceof EnumerationAxis) {
				Object value = ((EnumerationAxis<?>)axis).getEnumerationMemberAtIndex(s);
				if(value != null) {
					lbl = value.toString();
				}
			}
			Point2D segmentPoint = getSegmentPoint(segments, s, outer.getX() - center.getX(),
					outer.getY() - center.getY());
			drawSegmentIndicator(gc, lbl, aStyle, centerX, centerY, segmentPoint);
		}
	}

	/** Draws the indicator on the axis. */
	private void drawSegmentIndicator(GraphicsContext gc, String lbl, AxisStyle axisStyle,
			double centerX, double centerY, Point2D segment) {
		FontStyle segmentStyle = axisStyle.getSegmentStyle();
		// segment indicator labels are always to the left of the axis
		double textWidth = segmentStyle.getTextBounds(lbl).getWidth();
		double posX = centerX + segment.getX() - textWidth - 4;
		double posY = centerY + segment.getY() + 4;
		segmentStyle.drawText(gc, lbl, posX, posY);
	}

	/** Draws the axis label at the correct position. */
	private void drawAxisLabel(GraphicsContext gc, AxisBase axis, AxisStyle style, double centerX,
			double centerY, double outerX, double outerY) {
		String axisLabel = axis.getName();
		FontStyle labelStyle = style.getLabelStyle();
		Dimension2D tb = labelStyle.getTextBounds(axisLabel);
		int offset = 20;
		if(outerX > centerX) {
			outerX += offset;
		} else if(outerX < centerX) {
			outerX = outerX - offset - tb.getWidth();
		} else {
			outerX = outerX - tb.getWidth() / 2;
		}
		if(outerY > centerY) {
			outerY = outerY + offset + tb.getHeight() / 2;
		} else if(outerY < centerY) {
			outerY = outerY - offset;
		}
		labelStyle.drawText(gc, axisLabel, outerX, outerY);
	}
}
