/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.control;

import org.fortiss.tooling.common.ui.javafx.style.FontStyle;
import org.fortiss.tooling.common.ui.javafx.style.LineStyle;
import org.fortiss.tooling.spiderchart.model.DataSeries;
import org.fortiss.tooling.spiderchart.model.SpiderChart;
import org.fortiss.tooling.spiderchart.style.ChartStyle;
import org.fortiss.tooling.spiderchart.style.LegendStyle;

import javafx.geometry.Dimension2D;
import javafx.scene.canvas.GraphicsContext;

/**
 * Class used for drawing the the spider chart legend.
 * 
 * @author hoelzl
 */
public final class SpiderChartLegendControl extends SpiderChartControlBase {
	/** The length of the line for each data series. */
	private static final double LINE_SIZE = 20;
	/** The space between line and text. */
	private static final double LINE_SPACE = 5;

	/** Constructor. */
	public SpiderChartLegendControl(SpiderChart chart, ChartStyle style) {
		super(chart, style);
	}

	/** Draws the provided graphics */
	public void draw(GraphicsContext gc) {
		if(!style.isShowLegend()) {
			return;
		}

		if(style.getLegendStyle().isVerticalLayout()) {
			drawVertical(gc);
		} else {
			drawHorizontal(gc);
		}
	}

	/** Draws the graphics horizontally. */
	public void drawHorizontal(GraphicsContext gc) {
		String labelText = chart.getLegendLabel();
		double margin = style.getLegendStyle().getMargin();

		double xDelta = 0;
		if(labelText != null && labelText.trim().length() > 0) {
			xDelta = drawHorizontalLabel(gc);
		}
		double y = getY() + margin;
		for(DataSeries data : chart.getDataSeries()) {
			String text = data.getName();
			FontStyle fontStyle = style.getDataSeriesStyle(data).getIndicatorLabelStyle();
			Dimension2D extent = fontStyle.getTextBounds(text);

			LineStyle l = style.getDataSeriesStyle(data).getLineStyle();
			double xl = getX() + xDelta;
			l.drawLine(gc, xl, y - extent.getHeight() / 4, xl + LINE_SIZE,
					y - extent.getHeight() / 4);
			xDelta = xDelta + LINE_SIZE + LINE_SPACE;
			double xs = getX() + xDelta;
			fontStyle.drawText(gc, text, xs, y);
			xDelta = xDelta + extent.getWidth() + margin;
		}
	}

	/** Draws the horizontal legend label and returns the offset. */
	private double drawHorizontalLabel(GraphicsContext gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		FontStyle labelStyle = legendStyle.getLabelStyle();
		double margin = legendStyle.getMargin();

		double textExtent = labelStyle.getTextBounds(labelText).getWidth();
		labelStyle.drawText(gc, labelText, getX() + margin, getY() + margin);
		return textExtent + 2 * margin;
	}

	/** Draws the graphics vertically. */
	public void drawVertical(GraphicsContext gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		double margin = legendStyle.getMargin();

		double yDelta = 0;
		if(labelText != null && labelText.trim().length() > 0) {
			yDelta = drawVerticalLabel(gc);
		}

		double textWidth = 0;
		double textHeight = 0;

		for(DataSeries data : chart.getDataSeries()) {
			Dimension2D extent = legendStyle.getLabelStyle().getTextBounds(data.getName());
			textWidth = Math.max(textWidth, extent.getWidth());
			textHeight = Math.max(textHeight, extent.getHeight());
		}

		int cnt = 0;
		for(DataSeries data : chart.getDataSeries()) {
			double dY = yDelta + cnt * (textHeight + margin);
			double dX = LINE_SIZE + 2 * margin;
			LineStyle l = style.getDataSeriesStyle(data).getLineStyle();
			double lY = dY + textHeight / 2;
			l.drawLine(gc, getX() + margin, getY() + lY, getX() + margin + LINE_SIZE, getY() + lY);
			style.getDataSeriesStyle(data).getIndicatorLabelStyle().drawText(gc, data.getName(),
					getX() + dX, getY() + dY);
			cnt++;
		}
	}

	/** Draws the header label for the vertical legend layout and returns the offset. */
	private double drawVerticalLabel(GraphicsContext gc) {
		String labelText = chart.getLegendLabel();
		LegendStyle legendStyle = style.getLegendStyle();
		FontStyle labelStyle = legendStyle.getLabelStyle();
		double margin = legendStyle.getMargin();
		Dimension2D textExtent = labelStyle.getTextBounds(labelText);
		double x = getX() + margin + (getWidth() - margin - textExtent.getWidth()) / 2;
		labelStyle.drawText(gc, labelText, x, getY() + margin);
		return textExtent.getHeight() + margin;
	}
}
