/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Apache License, Version 2.0, which is available at 
 * https://www.apache.org/licenses/LICENSE-2.0.
 *
 * SPDX-License-Identifier: Apache-2.0
 *******************************************************************************/
package org.fortiss.tooling.spiderchart.model;

import static java.util.Collections.unmodifiableList;

import java.util.ArrayList;
import java.util.List;

/**
 * Main class of the spider chart model.
 * 
 * @author hoelzl
 */
public final class SpiderChart {
	/** The axes contained in this chart. */
	private List<AxisBase> axes = new ArrayList<>();
	/** The data sets contained in this chart. */
	private List<DataSeries> dataSeries = new ArrayList<DataSeries>();
	/** The title of the spider chart. */
	private String title;
	/** The label of the legend. */
	private String legendLabel;

	/** Adds the given data series. */
	public void addData(DataSeries d) {
		dataSeries.add(d);
	}

	/** Adds the given axis. */
	public void addAxis(AxisBase a) {
		axes.add(a);
	}

	/** Sets title. */
	public void setTitle(String title) {
		this.title = title;
	}

	/** Sets legendLabel. */
	public void setLegendLabel(String legendLabel) {
		this.legendLabel = legendLabel;
	}

	/**
	 * Validates the chart by checking if all data series provide a non-null value for each axis and
	 * that this value is within the axis minimum an maximum values..
	 */
	public boolean checkDataValidity() {
		for(DataSeries d : dataSeries) {
			for(AxisBase a : axes) {
				if(d.getAxisValue(a) == null) {
					return false;
				}
				if(!a.isValidValue(d.getAxisValue(a))) {
					return false;
				}
			}
		}
		return true;
	}

	/** Returns the unmodifiable list of axes. */
	public List<AxisBase> getAxes() {
		return unmodifiableList(axes);
	}

	/** Returns the unmodifiable list of data series. */
	public List<DataSeries> getDataSeries() {
		return unmodifiableList(dataSeries);
	}

	/** Returns the legend label. */
	public String getLegendLabel() {
		return legendLabel;
	}

	/** Returns title. */
	public String getTitle() {
		return title;
	}

	/**
	 * Moves the given {@link DataSeries} to the end of the list (causing drawing to be performed
	 * last).
	 */
	public void moveSeriesToEnd(DataSeries ds) {
		dataSeries.remove(ds);
		dataSeries.add(ds);
	}
}
