/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.editor.errorhandler;

import org.eclipse.core.resources.IResource;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.jface.layout.GridLayoutFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.forms.editor.FormEditor;

import org.conqat.lib.commons.string.StringUtils;
import org.conqat.ide.commons.ui.ECCSMCommonsImage;
import org.conqat.ide.commons.ui.swt.SWTUtils;

/**
 * This class deals with errors which occur during the initialization of an
 * editor. It simplifies checking for common problems and allows to display an
 * error page in case an error occurred.
 * <p>
 * See one of the users of this class for how to apply this to an editor.
 * 
 * @author hummelb
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: 247B197755E7AC79B0D4E9C88AA96B9A
 */
public class EditorErrorHandler {

	/** The error message. If this is null, no error occurred. */
	private String errorMessage;

	/**
	 * A string that details the error message (may be null if no details are
	 * available).
	 */
	private String errorDetails;

	/** Sets the error message. */
	public void setErrorMessage(String errorMessage, String errorDetails) {
		this.errorMessage = errorMessage;
		this.errorDetails = errorDetails;
	}

	/**
	 * Sets the error message from an exception. This uses the message for the
	 * main message and the full stack trace for the details.
	 */
	public void setErrorMessage(Throwable t) {
		setErrorMessage(t.getMessage(), StringUtils.obtainStackTrace(t));
	}

	/**
	 * Checks a resource for several commonly used properties:
	 * <ul>
	 * <li>null check</li>
	 * <li>Existence of the resource</li>
	 * <li>Out of sync check</li>
	 * </ul>
	 * If any of the checks fails, the error message is updated and false is
	 * returned (otherwise true).
	 */
	public boolean checkResource(IResource resource) {

		if (resource == null) {
			setErrorMessage("Invalid editor input!", null);
			return false;
		}

		if (!resource.exists()) {
			setErrorMessage("Resource " + resource.getName()
					+ " does not exist!", null);
			return false;
		}

		if (!resource.isSynchronized(0)) {
			setErrorMessage("Resource " + resource.getName()
					+ " is out of sync. "
					+ "Perform a refresh on the file and reopen the editor.",
					null);
			return false;
		}

		return true;
	}

	/** Returns whether there is an error message set. */
	public boolean hasError() {
		return errorMessage != null;
	}

	/**
	 * This method displays an error page on the given parent composite if an
	 * error message is set. Otherwise nothing happens. It expects that this
	 * composite can be used exclusively.
	 * 
	 * @return <code>true</code> if there was an error and thus the error page
	 *         has been added to the composite. <code>false</code> otherwise.
	 */
	public boolean displayErrorPage(Composite parent) {
		if (!hasError()) {
			return false;
		}

		Composite main = new Composite(parent, SWT.NONE);

		Color white = Display.getCurrent().getSystemColor(SWT.COLOR_WHITE);
		main.setBackground(white);

		GridLayoutFactory.swtDefaults().extendedMargins(10, 10, 10, 10)
				.numColumns(2).applyTo(main);

		Label icon = new Label(main, SWT.NONE);
		icon.setBackground(white);
		icon.setImage(ECCSMCommonsImage.ERROR.getImage());
		SWTUtils.layout(icon);

		Label label = new Label(main, SWT.NONE);
		label.setBackground(white);
		label.setText("Error opening the editor:");
		SWTUtils.layout(label);

		GridDataFactory gdf = GridDataFactory.swtDefaults().grab(true, true)
				.align(SWT.FILL, SWT.FILL).span(2, 1);

		Text message = new Text(main, SWT.BORDER | SWT.READ_ONLY | SWT.MULTI
				| SWT.WRAP | SWT.V_SCROLL);
		message.setText(errorMessage);
		gdf.hint(-1, 50).applyTo(message);
		message.setBackground(white);

		if (errorDetails == null) {
			return true;
		}

		Text detail = new Text(main, SWT.BORDER | SWT.READ_ONLY | SWT.MULTI
				| SWT.WRAP | SWT.V_SCROLL);
		detail.setText(errorDetails);
		gdf.indent(0, 20).hint(-1, 100).applyTo(detail);
		detail.setBackground(white);

		return true;
	}

	/**
	 * This method displays an error page in the given form editor if an error
	 * message is set. Otherwise nothing happens. It expects that this form
	 * editor can be used exclusively.
	 * 
	 * @param parent
	 *            the parent to be used for creation of new controls. Typically
	 *            this would be formEditor.getContainer().
	 * 
	 * @return true if there was an error and thus the error page has been added
	 *         to the editor. False otherwise.
	 */
	public boolean displayErrorPage(FormEditor formEditor, Composite parent) {
		if (!hasError()) {
			return false;
		}

		Composite composite = new Composite(parent, SWT.NONE);
		composite.setLayout(new FillLayout());
		displayErrorPage(composite);
		formEditor.addPage(composite);

		return true;
	}
}